"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/1/27 9:55
@DependencyLibrary：无
@MainFunction：无
@FileDoc：
    test_serializer.py
    测试文件
@ChangeHistory:
    datetime action why
    example:
    2021/1/27 9:55 change 'Fix bug'

"""
import asyncio
import datetime
import unittest
from collections import OrderedDict
from copy import deepcopy
from decimal import Decimal

from sanic_rest_framework.fields import (
    CharField, IntegerField, FloatField, DateField, TimeField, DecimalField,
    DateTimeField, BooleanField, ChoiceField, SerializerMethodField)
from sanic_rest_framework.serializers import Serializer

from tortoise.contrib.test import initializer, finalizer, TestCase
from sanic_rest_framework.test.models import UserModel, AddressModel, SchoolModel, StudentModel, ClassRoomModel
from sanic_rest_framework.exceptions import ValidationError


class AddressSerializer(Serializer):
    phone = CharField(max_length=11, required=True)
    address = CharField(max_length=100)
    house_number = CharField(max_length=100)


class UserSerializer(Serializer):
    name = CharField(max_length=8, required=True)
    birthday = DateField()
    phone = CharField(max_length=11)
    balance = DecimalField(13, 3)
    address = AddressSerializer(many=True)


class SchoolSerializer(Serializer):
    name = CharField(max_length=12)
    address = AddressSerializer()


class StudentSerializer(Serializer):
    name = CharField(max_length=12)
    # class_room = ClassRoomSerializer()


class ClassRoomSerializer(Serializer):
    room_number = CharField(max_length=12)
    student_count = IntegerField()
    students = StudentSerializer(many=True)


class M2OClassRoomSerializer(Serializer):
    room_number = CharField(max_length=12)
    student_count = IntegerField()


class M2OStudentSerializer(Serializer):
    name = CharField(max_length=12)
    class_room = M2OClassRoomSerializer()


initializer(['sanic_rest_framework.test.models', ],
            # db_url="sqlite://./db.sqlite",
            loop=asyncio.get_event_loop())


class TestOrdinarySerializer(TestCase):
    def setUp(self) -> None:
        self.mapping_data = {
            'phone': '17674707036',
            'address': '长沙市IFS',
            'house_number': '67L'
        }
        self.res = OrderedDict(self.mapping_data)

    async def test_serializer(self):
        """测试序列化"""
        addr_model = AddressModel(**self.mapping_data)
        await addr_model.save()
        as_map = AddressSerializer(instance=self.mapping_data)
        as_model = AddressSerializer(instance=await AddressModel.get(pk=1))
        self.assertEqual(await as_map.data, self.res)
        self.assertEqual(await as_model.data, self.res)

    async def test_deserializer(self):
        """测试反序列化"""

        addr_model = AddressModel(**self.mapping_data)
        await addr_model.save()
        as_map = AddressSerializer(data=self.mapping_data)
        as_model = AddressSerializer(data=await AddressModel.get(pk=1))
        self.assertIs(as_map.is_valid(), True)
        self.assertEqual(as_map.validated_data, self.res)

        self.assertIs(as_model.is_valid(), False)
        print(as_model.errors)
        # print(as_model.validated_data)

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()


class TestM2MNestedSerializer(TestCase):
    """测试多对多嵌套"""

    def setUp(self) -> None:
        self.mapping_data = {
            'name': '老四',
            'birthday': '2010-11-16',
            'phone': '17674707037',
            'balance': '99.009',
            'address': [{
                'phone': '17674707036',
                'address': '长沙市IFS',
                'house_number': '67L'
            }, {
                'phone': '17674707037',
                'address': '长沙市HPT',
                'house_number': '4L'
            }]
        }
        self.des_res = OrderedDict({'name': '老四', 'birthday': datetime.date(2010, 11, 16),
                                    'phone': '17674707037', 'balance': Decimal('99.009'),
                                    'address': [
                                        OrderedDict({'phone': '17674707036', 'address': '长沙市IFS', 'house_number': '67L'}),
                                        OrderedDict({'phone': '17674707037', 'address': '长沙市HPT', 'house_number': '4L'})]})
        self.s_res = OrderedDict({'name': '老四', 'birthday': '2010-11-16',
                                    'phone': '17674707037', 'balance': Decimal('99.009'),
                                    'address': [
                                        OrderedDict({'phone': '17674707036', 'address': '长沙市IFS', 'house_number': '67L'}),
                                        OrderedDict({'phone': '17674707037', 'address': '长沙市HPT', 'house_number': '4L'})]})
        self.res = OrderedDict(self.mapping_data)

    async def test_serializer(self):
        """测试序列化"""
        user_info = deepcopy(self.mapping_data)
        address = user_info.pop('address')
        user = UserModel(**user_info)
        await user.save()
        for i in address:
            addr = AddressModel(**i)
            await addr.save()
            await user.address.add(addr)

        us_map = UserSerializer(instance=self.mapping_data)
        us_model = UserSerializer(instance=await UserModel.get(pk=1))
        print(await us_model.data)
        self.assertEqual(await us_map.data, self.s_res)
        self.assertEqual(await us_model.data, self.s_res)

    async def test_deserializer(self):
        """测试反序列化"""
        user_info = deepcopy(self.mapping_data)
        address = user_info.pop('address')
        user = UserModel(**user_info)
        await user.save()
        for i in address:
            addr = AddressModel(**i)
            await addr.save()
            await user.address.add(addr)
        us_map = UserSerializer(data=self.mapping_data)
        us_model = UserSerializer(data=await UserModel.get(pk=1))
        self.assertIs(us_map.is_valid(), True)
        self.assertEqual(us_map.validated_data, self.des_res)

        self.assertIs(us_model.is_valid(), False)
        print(us_model.errors)

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()


class TestO2OSerializer(TestCase):
    """测试一对一序列化"""

    def setUp(self) -> None:
        self.mapping_data = {
            'name': '老四',
            'address': {
                'phone': '17674707036',
                'address': '长沙市IFS',
                'house_number': '67L'
            }
        }
        self.res = OrderedDict(self.mapping_data)

    async def test_serializer(self):
        """测试序列化"""
        school_info = deepcopy(self.mapping_data)
        address = school_info.pop('address')
        address_model = await AddressModel.create(**address)
        school_model = await SchoolModel.create(address=address_model, **school_info)

        ss_map = SchoolSerializer(instance=self.mapping_data)
        us_model = SchoolSerializer(instance=await SchoolModel.get(pk=1))
        self.assertEqual(await ss_map.data, self.res)
        self.assertEqual(await us_model.data, self.res)

    async def test_deserializer(self):
        """测试反序列化"""
        school_info = deepcopy(self.mapping_data)
        address = school_info.pop('address')
        address_model = await AddressModel.create(**address)
        school_model = await SchoolModel.create(address=address_model, **school_info)

        ss_map = SchoolSerializer(data=self.mapping_data)
        ss_model = SchoolSerializer(data=await SchoolModel.get(pk=1))
        self.assertIs(ss_map.is_valid(), True)
        self.assertEqual(ss_map.validated_data, self.res)

        self.assertIs(ss_model.is_valid(), False)
        print(ss_model.errors)

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()


class TestO2MSerializer(TestCase):
    """测试一对多序列化器"""

    def setUp(self) -> None:
        self.mapping_data = {
            'room_number': '1024',
            'student_count': 80,
            'students': [{'name': '刘文静'}, {'name': '马冬梅'}, {'name': '光明'}]
        }
        self.res = OrderedDict(self.mapping_data)

    async def test_serializer(self):
        """测试序列化"""

        class_room_info = deepcopy(self.mapping_data)
        students = class_room_info.pop('students')
        class_model = await ClassRoomModel.create(**class_room_info)
        await StudentModel.bulk_create([StudentModel(class_room=class_model, **student) for student in students])

        cs_map = ClassRoomSerializer(instance=self.mapping_data)
        cs_model = ClassRoomSerializer(instance=class_model)
        print(await cs_map.data)
        print(await cs_model.data)
        self.assertEqual(await cs_map.data, self.res)
        self.assertEqual(await cs_model.data, self.res)

    async def test_deserializer(self):
        """测试反序列化"""
        class_room_info = deepcopy(self.mapping_data)
        students = class_room_info.pop('students')
        class_model = await ClassRoomModel.create(**class_room_info)
        await StudentModel.bulk_create([StudentModel(class_room=class_model, **student) for student in students])

        cs_map = ClassRoomSerializer(data=self.mapping_data)
        cs_model = ClassRoomSerializer(data=class_model)
        self.assertIs(cs_map.is_valid(), True)
        self.assertEqual(cs_map.validated_data, self.res)

        self.assertIs(cs_model.is_valid(), False)
        print(cs_model.errors)

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()


class TestM2OSerializer(TestCase):
    """测试多对一序列化器"""

    def setUp(self) -> None:
        self.mapping_data = [
            {
                'name': '马冬梅',
                'class_room': {
                    'room_number': '1024',
                    'student_count': 40
                }
            }, {
                'name': '光明',
                'class_room': {
                    'room_number': '1024',
                    'student_count': 40
                }
            }, {
                'name': '李焕英',
                'class_room': {
                    'room_number': '1024',
                    'student_count': 40
                }
            },

        ]
        self.res_dict = {
            '马冬梅': {
                'name': '马冬梅',
                'class_room': {
                    'room_number': '1024',
                    'student_count': 40
                }
            }, '光明': {
                'name': '光明',
                'class_room': {
                    'room_number': '1024',
                    'student_count': 40
                }
            }, '李焕英': {
                'name': '李焕英',
                'class_room': {
                    'room_number': '1024',
                    'student_count': 40
                }
            },
        }
        self.res = OrderedDict(self.mapping_data)

    async def test_serializer(self):
        """测试序列化"""

        student_model_list = []
        students = deepcopy(self.mapping_data)
        # 初始化数据库数据
        for student_info in students:
            class_room_info = student_info.pop('class_room')
            class_room_model, status = await ClassRoomModel.get_or_create(**class_room_info)
            student_model = await StudentModel.create(class_room=class_room_model, **student_info)
            student_model_list.append(student_model)

        # 验证模型类型的数据是否正常
        for student_model in student_model_list:
            m2oss_model = M2OStudentSerializer(instance=student_model)
            self.assertEqual(await m2oss_model.data, OrderedDict(self.res_dict[student_model.name]))

        # 验证字段类型的数据是否正常
        students = deepcopy(self.mapping_data)
        for student_info in students:
            m2oss_map = M2OStudentSerializer(instance=student_info)
            self.assertEqual(await m2oss_map.data, OrderedDict(self.res_dict[student_info['name']]))

    async def test_deserializer(self):
        """测试反序列化"""
        student_model_list = []
        students = deepcopy(self.mapping_data)
        # 初始化数据库数据
        for student_info in students:
            class_room_info = student_info.pop('class_room')
            class_room_model, status = await ClassRoomModel.get_or_create(**class_room_info)
            student_model = await StudentModel.create(class_room=class_room_model, **student_info)
            student_model_list.append(student_model)

        # 验证模型类型的数据是否正常
        for student_model in student_model_list:
            m2oss_model = M2OStudentSerializer(data=student_model)
            self.assertIs(m2oss_model.is_valid(), False)
            print(m2oss_model.errors)

        # 验证字段类型的数据是否正常
        students = deepcopy(self.mapping_data)
        for student_info in students:
            m2oss_map = M2OStudentSerializer(data=student_info)
            self.assertIs(m2oss_map.is_valid(), True)
            self.assertEqual(m2oss_map.validated_data, OrderedDict(self.res_dict[student_info['name']]))

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()


if __name__ == '__main__':
    unittest.main()



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/19 15:44
@DependencyLibrary:
@MainFunction：
@FileDoc:
    views.py
    基础视图文件
"""
from sanic.response import json
from tortoise.queryset import QuerySet
from sanic.log import logger

from sanic_rest_framework.constant import ALL_METHOD
from sanic_rest_framework.exceptions import APIException, ValidationError
from sanic_rest_framework.filters import SearchFilter
from sanic_rest_framework.paginations import GeneralPagination
from sanic_rest_framework.status import RuleStatus, HttpStatus
from simplejson import dumps


class BaseAPIView:
    """基础API视图"""
    authentication_classes = ()
    permission_classes = ()

    async def dispatch(self, request, *args, **kwargs):
        """分发路由"""
        request.user = None
        method = request.method
        if method not in self.licensed_methods:
            return self.json_response(msg='发生错误：未找到%s方法' % method, status=RuleStatus.STATUS_0_FAIL,
                                      http_status=HttpStatus.HTTP_405_METHOD_NOT_ALLOWED)
        handler = getattr(self, method.lower(), None)

        try:
            self.initial(request, *args, **kwargs)
            response = await handler(request, *args, **kwargs)
        except APIException as exc:
            response = self.handle_exception(exc)
        except ValidationError as exc:
            response = self.error_json_response(exc.message_dict, '数据验证失败')
        except AssertionError as exc:
            raise exc
        except Exception as exc:
            logger.error('--捕获未知错误--', exc)
            response = self.handle_uncaught_exception(exc)
        return response

    @classmethod
    def as_view(cls, methods=None, *class_args, **class_kwargs):

        # 许可的方法
        if methods is None:
            methods = ALL_METHOD

        # 返回的响应方法闭包
        def view(request, *args, **kwargs):
            self = view.base_class(*class_args, **class_kwargs)
            self.licensed_methods = methods
            self.request = request
            self.args = args
            self.kwargs = kwargs
            return self.dispatch(request, *args, **kwargs)

        view.base_class = cls
        view.API_DOC_CONFIG = class_kwargs.get('API_DOC_CONFIG')  # 未来的API文档配置属性+
        view.__doc__ = cls.__doc__
        view.__module__ = cls.__module__
        view.__name__ = cls.__name__
        return view

    def handle_exception(self, exc: APIException):
        return self.json_response(**exc.response_data())

    def handle_uncaught_exception(self, exc):
        """处理未知错误"""
        message = '{}:{}'.format(exc.__class__.__name__, '|'.join(exc.args))
        return self.json_response(msg=message, status=RuleStatus.STATUS_0_FAIL,
                                  http_status=HttpStatus.HTTP_500_INTERNAL_SERVER_ERROR)

    def json_response(self, data=None, msg="OK", status=RuleStatus.STATUS_1_SUCCESS,
                      http_status=HttpStatus.HTTP_200_OK):
        """
        Json 相应体
        :param data: 返回的数据主题
        :param msg: 前台提示字符串
        :param status: 前台约定状态，供前台判断是否成功
        :param http_status: Http响应数据
        :return:
        """
        if data is None:
            data = {}
        response_body = {
            'data': data,
            'message': msg,
            'status': status
        }
        return json(body=response_body, status=http_status, dumps=dumps)

    def success_json_response(self, data=None, msg="Success", **kwargs):
        """
        快捷的成功的json响应体
        :param data: 返回的数据主题
        :param msg: 前台提示字符串
        :return: json
        """
        status = kwargs.pop('status', RuleStatus.STATUS_1_SUCCESS)
        http_status = kwargs.pop('http_status', HttpStatus.HTTP_200_OK)
        return self.json_response(data=data, msg=msg, status=status, http_status=http_status)

    def error_json_response(self, data=None, msg="Fail", **kwargs):
        """
        快捷的失败的json响应体
        :param data: 返回的数据主题
        :param msg: 前台提示字符串
        :return: json
        """
        status = kwargs.pop('status', RuleStatus.STATUS_0_FAIL)
        http_status = kwargs.pop('http_status', HttpStatus.HTTP_400_BAD_REQUEST)
        return self.json_response(data=data, msg=msg, status=status, http_status=http_status)

    async def get_object(self):
        """
        返回视图显示的对象。
        如果您需要提供非标准的内容，则可能要覆盖此设置
        queryset查找。
        """
        queryset = self.filter_queryset(self.get_queryset())

        lookup_field = self.lookup_field

        assert lookup_field in self.kwargs, (
                '%s 不存在于 %s 的 Url配置中的关键词内 ' %
                (lookup_field, self.__class__.__name__,)
        )

        filter_kwargs = {lookup_field: self.kwargs[lookup_field]}
        obj = await queryset.get_or_none(**filter_kwargs)
        if obj is None:
            raise APIException('不存在%s为%s的数据' % (lookup_field, self.kwargs[lookup_field]), http_status=HttpStatus.HTTP_200_OK)

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    def get_authenticators(self):
        """
        实例化并返回此视图可以使用的身份验证器列表
        """
        return [auth() for auth in self.authentication_classes]

    def check_authentication(self, request):
        """
        检查权限 查看是否拥有权限，并在此处为Request.User 赋值
        :param request: 请求
        :return:
        """
        for authenticators in self.get_authenticators():
            request.user = authenticators.authenticate(request, request.user)

    def get_permissions(self):
        """
        实例化并返回此视图所需的权限列表
        """
        return [permission() for permission in self.permission_classes]

    def check_permissions(self, request):
        """
        检查是否应允许该请求，如果不允许该请求，
        则在 has_permission 中引发一个适当的异常。
        :param request: 当前请求
        :return:
        """
        for permission in self.get_permissions():
            permission.has_permission(request, self)

    def check_object_permissions(self, request, obj):
        """
        检查是否应允许给定对象的请求, 如果不允许该请求，
        则在 has_object_permission 中引发一个适当的异常。
            常用于 get_object() 方法
        :param request: 当前请求
        :param obj: 需要鉴权的模型对象
        :return:
        """
        for permission in self.get_permissions():
            permission.has_object_permission(request, self, obj)

    def check_throttles(self, request):
        """
        检查范围频率。
        则引发一个 APIException 异常。
        :param request:
        :return:
        """
        pass

    def initial(self, request, *args, **kwargs):
        """
        在请求分发之前执行初始化操作，用于检查权限及检查基础内容
        """
        self.check_authentication(request)
        self.check_permissions(request)
        self.check_throttles(request)


class APIView(BaseAPIView):
    queryset = None
    lookup_field = 'pk'
    serializer_class = None
    pagination_class = None
    filter_class = SearchFilter
    search_fields = None

    def get_queryset(self):
        assert self.queryset is not None, (
                "'%s'应该包含一个'queryset'属性，"
                "或重写`get_queryset()`方法。"
                % self.__class__.__name__
        )

        queryset = self.queryset
        if isinstance(queryset, QuerySet):
            queryset = queryset.all()
        return queryset

    def filter_queryset(self, queryset):
        return self.filter_class().filter_queryset(self.request, queryset, self)

    def get_serializer(self, *args, **kwargs):
        """
        返回应该用于验证和验证的序列化程序实例
        对输入进行反序列化，并对输出进行序列化。
        """
        serializer_class = self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self):
        """
        返回用于序列化器的类。
        默认使用`self.serializer_class`。

        如果您需要提供其他信息，则可能要覆盖此设置
        序列化取决于传入的请求。

        （例如，管理员获得完整的序列化，其他获得基本的序列化）
        """
        assert self.serializer_class is not None, (
                "'%s' should either include a `serializer_class` attribute, "
                "or override the `get_serializer_class()` method."
                % self.__class__.__name__
        )
        return self.serializer_class

    def get_serializer_context(self):
        """
        提供给序列化程序类的额外上下文。
        """
        return {
            'request': self.request,
            'view': self
        }

    @property
    def paginator(self):
        """
        与视图关联的分页器实例，或“None”。
        """
        if not hasattr(self, '_paginator'):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    async def paginate_queryset(self, queryset):
        """
        返回单页结果，如果禁用了分页，则返回“无”。
        """
        if self.paginator is None:
            return None
        return await self.paginator.paginate_queryset(queryset, self.request, view=self)

    def get_paginated_response(self, data):
        """
        返回给定输出数据的分页样式`Response`对象。
        """
        assert self.paginator is not None
        return self.paginator.response(self.request, data)


class ListModelMixin:
    """
    适用于输出列表类型数据
    """
    pagination_class = GeneralPagination
    detail = False

    async def get(self, request, *args, **kwargs):
        return await self.list(request, *args, **kwargs)

    async def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = await self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            data = self.get_paginated_response(await serializer.data)
            return self.success_json_response(data=data)

        serializer = self.get_serializer(queryset, many=True)
        return self.success_json_response(data=await serializer.data)


class CreateModelMixin:
    """
    适用于快速创建内容
    占用 post 方法
    """

    async def post(self, request, *args, **kwargs):
        return await self.create(request, *args, **kwargs)

    async def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        await serializer.is_valid(raise_exception=True)
        await self.perform_create(serializer)
        return self.success_json_response(data=await serializer.data, http_status=HttpStatus.HTTP_201_CREATED)

    async def perform_create(self, serializer):
        await serializer.save()


class DetailModelMixin:
    """
    适用于查询指定PK的内容
    """
    detail = True

    async def get(self, request, *args, **kwargs):
        return await self.detail(request, *args, **kwargs)

    async def detail(self, request, *args, **kwargs):
        instance = await self.get_object()
        serializer = self.get_serializer(instance)
        return self.success_json_response(data=await serializer.data)


class UpdateModelMixin:
    """
    适用于快速创建更新操作
    """

    async def put(self, request, *args, **kwargs):
        return await self.update(request, *args, **kwargs)

    async def patch(self, request, *args, **kwargs):
        return await self.partial_update(request, *args, **kwargs)

    async def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = await self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        await serializer.is_valid(raise_exception=True)
        await self.perform_update(serializer)

        # if getattr(instance, '_prefetched_objects_cache', None):
        #     instance._prefetched_objects_cache = {}

        return self.success_json_response(data=await serializer.data)

    async def perform_update(self, serializer):
        await serializer.save()

    async def partial_update(self, request, *args, **kwargs):
        kwargs['partial'] = True
        return await self.update(request, *args, **kwargs)


class DestroyModelMixin:
    """
    用于快速删除
    """

    async def delete(self, request, *args, **kwargs):
        return await self.destroy(request, *args, **kwargs)

    async def destroy(self, request, *args, **kwargs):
        instance = await self.get_object()
        await self.perform_destroy(instance)
        return self.success_json_response(status=HttpStatus.HTTP_204_NO_CONTENT)

    async def perform_destroy(self, instance):
        await instance.delete()



from datetime import date
from enum import Enum, IntEnum

from tortoise import fields
from tortoise import Model
from tortoise.fields import ForeignKeyRelation, ReverseRelation


class Enum1(IntEnum):
    OK = 1
    BAD = 2


class CharEnum(Enum):
    OK = '好'
    BAD = '坏'


class TestModel(Model):
    char_field = fields.CharField(max_length=8, null=True)
    float_field = fields.FloatField()
    date_field = fields.DateField()
    int_field = fields.IntField()
    decimal_field = fields.DecimalField(max_digits=13, decimal_places=3)
    datetime_field = fields.DatetimeField()
    int_enum_field = fields.IntEnumField(enum_type=Enum1)
    char_enum_field = fields.CharEnumField(enum_type=CharEnum)
    boolean_field = fields.BooleanField()
    small_int_field = fields.SmallIntField()
    big_int_field = fields.BigIntField()
    text_field = fields.TextField()
    json_field = fields.JSONField()
    uuid_field = fields.UUIDField()
    one_to_many: ForeignKeyRelation["TestManyToOneModel"] = fields.ForeignKeyField('models.TestManyToOneModel', related_name='many_to_one', null=True)
    one_to_one: fields.OneToOneRelation["TestOneToOneModel"] = fields.OneToOneField('models.TestOneToOneModel', related_name='one_to_one', null=True)
    many_to_many: fields.ManyToManyRelation["TestModel"] = fields.ManyToManyField("models.ManyToManyModel", related_name="many_2_many", through="many_many", null=True)


class TestOneTOManyModel(Model):
    one_to_many: ForeignKeyRelation["TestModel"] = fields.ForeignKeyField('models.TestModel', related_name='many_to_one')


class TestManyToOneModel(Model):
    many_to_one: fields.ReverseRelation["TestModel"]


class TestOneToOneModel(Model):
    name = fields.CharField(max_length=8, null=True)


class ManyToManyModel(Model):
    many_2_many: fields.ManyToManyRelation["TestModel"]


class UserModel(Model):
    name = fields.CharField(max_length=8, null=False)
    birthday = fields.DateField()
    phone = fields.CharField(max_length=11)
    balance = fields.DecimalField(13, 3)
    address: fields.ManyToManyRelation["AddressModel"] = fields.ManyToManyField(
        'models.AddressModel', through='user2address', related_name='user', null=True)


class AddressModel(Model):
    phone = fields.CharField(12, null=False)
    address = fields.CharField(100)
    house_number = fields.CharField(100)
    # user: fields.ManyToManyRelation[UserModel]


class SchoolModel(Model):
    name = fields.CharField(12)
    address: fields.OneToOneRelation["AddressModel"] = fields.OneToOneField("models.AddressModel", 'school')


class ClassRoomModel(Model):
    room_number = fields.CharField(18)
    student_count = fields.IntField()
    students: ReverseRelation['StudentModel']


class StudentModel(Model):
    name = fields.CharField(max_length=12, null=False)
    class_room: ForeignKeyRelation["ClassRoomModel"] = fields.ForeignKeyField('models.ClassRoomModel', 'students')


class DateSeriesModel(Model):
    name = fields.TimeDeltaField()



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/20 20:03
@DependencyLibrary: 无
@MainFunction：无
@FileDoc:
    exceptions.py
    序列化器文件
"""
from typing import Mapping

from sanic.response import json

from sanic_rest_framework.status import HttpStatus, RuleStatus


class ValidationError(Exception):
    """验证器通用错误类 发生错误即抛出此类"""

    def __init__(self, message, code=None, params=None):
        super().__init__(message, code, params)
        if isinstance(message, ValidationError):
            if hasattr(message, 'error_dict'):
                message = message.error_dict
            elif not hasattr(message, 'message'):
                message = message.error_list
            else:
                message, code, params = message.message, message.code, message.params
        if isinstance(message, dict):
            self.error_dict = {}
            for field, msg in message.items():
                if not isinstance(msg, ValidationError):
                    msg = ValidationError(msg)
                if hasattr(msg, 'error_dict'):
                    self.error_dict[field] = [msg.error_dict]
                else:
                    self.error_dict[field] = msg.error_list
        elif isinstance(message, list):
            self.error_list = []
            for message in message:
                if not isinstance(message, ValidationError):
                    message = ValidationError(message)
                if hasattr(message, 'error_dict'):
                    self.error_list.extend(sum(message.error_dict.values(), []))
                else:
                    self.error_list.extend(message.error_list)
        else:
            self.message = message
            self.code = code
            self.params = params
            self.error_list = [self]

    @property
    def message_dict(self):
        getattr(self, 'error_dict')
        return dict(self)

    @property
    def messages(self):
        if hasattr(self, 'error_dict'):
            return sum(dict(self).values(), [])
        return list(self)

    def update_error_dict(self, error_dict):
        if hasattr(self, 'error_dict'):
            for field, error_list in self.error_dict.items():
                error_dict.setdefault(field, []).extend(error_list)
        else:
            error_dict.setdefault('__all__', []).extend(self.error_list)
        return error_dict

    def __iter__(self):
        if hasattr(self, 'error_dict'):
            for field, errors in self.error_dict.items():
                yield field, list(ValidationError(errors))
        else:
            for error in self.error_list:
                message = error.message
                if error.params:
                    message %= error.params
                yield str(message)

    def __str__(self):
        if hasattr(self, 'error_dict'):
            return repr(dict(self))
        return repr(list(self))

    def __repr__(self):
        return 'ValidationError(%s)' % self

    def __eq__(self, other):
        if not isinstance(other, ValidationError):
            return NotImplemented
        return hash(self) == hash(other)


class ValidatorAssertError(Exception):
    pass


class APIException(Exception):
    def __init__(self, message, status=RuleStatus.STATUS_0_FAIL, http_status=HttpStatus.HTTP_500_INTERNAL_SERVER_ERROR,
                 *args, **kwargs):
        self.message = message
        self.status = status
        self.http_status = http_status

    def response_data(self):
        return {
            'msg': self.message,
            'status': self.status,
            'http_status': self.http_status
        }



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/3/15 19:18
@DependencyLibrary: 无
@MainFunction：无
@FileDoc:
    __init__.py.py
    文件描述
@ChangeHistory:
    datetime action why
    example:
    2021/2/4 14:35 change 'Fix bug'
        
"""



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/20 13:20
@DependencyLibrary:
@MainFunction：
@FileDoc:
    serializers.py
    序列化器文件
"""
import copy
import inspect
from asyncio import coroutine
from collections import OrderedDict
from typing import Any, Mapping, Coroutine

from tortoise import models, Model
from tortoise.queryset import QuerySet
from tortoise import fields as tortoise_fields

from sanic_rest_framework.fields import (
    empty, SkipField,
    Field, CharField, IntegerField, FloatField, DecimalField, BooleanField, DateTimeField, DateField, TimeField, ChoiceField, SerializerMethodField
)
from sanic_rest_framework.converter import ModelConverter
from .exceptions import ValidationError
from .helpers import BindingDict

LIST_SERIALIZER_KWARGS = (
    'read_only',
    'write_only',
    'required',
    'allow_null',
    'default',
    'source',
    'validators',
    'error_messages',
    'label',
    'description',
    'instance',
    'data',
    'partial'
)
ALL_FIELDS = '__all__'


def set_value(dictionary, keys, value):
    """
    set_value({'a': 1}, [], {'b': 2}) -> {'a': 1, 'b': 2}
    set_value({'a': 1}, ['x'], 2) -> {'a': 1, 'x': 2}
    set_value({'a': 1}, ['x', 'y'], 2) -> {'a': 1, 'x': {'y': 2}}
    """
    if not keys:
        dictionary.update(value)
        return

    for key in keys[:-1]:
        if key not in dictionary:
            dictionary[key] = {}
        dictionary = dictionary[key]

    dictionary[keys[-1]] = value


class BaseSerializer(Field):
    """序列化器
    .instance ->
        .get_internal_value ->
            .internal_to_external() ->
                .data

    .install_data -> .get_external_value -> .external_to_internal() -> .validated_data

    """

    def __init__(self, instance=None, data=empty, **kwargs):
        self.instance = instance
        if data is not empty:
            self.initial_data = data
        self.partial = kwargs.pop('partial', False)
        self._context = kwargs.pop('context', {})
        kwargs.pop('many', None)
        super().__init__(**kwargs)

    def __new__(cls, *args, **kwargs):
        if kwargs.pop('many', False):
            return cls.many_init(*args, **kwargs)
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def many_init(cls, *args, **kwargs):
        """"""
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
            key: value for key, value in kwargs.items()
            if key in LIST_SERIALIZER_KWARGS
        })
        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(meta, 'list_serializer_class', ListSerializer)
        return list_serializer_class(*args, **list_kwargs)

    async def external_to_internal(self, data: Any) -> Any:
        """对数据进行序列化转换并返回"""
        raise NotImplementedError(
            '{cls}类的 .external_to_internal 方法必须重写'.format(cls=self.__class__.__name__, )
        )

    async def internal_to_external(self, data: Any) -> Any:
        """对数据进行反序列化转换并返回"""
        raise NotImplementedError(
            '{cls}类的 .external_to_internal 方法必须重写'.format(cls=self.__class__.__name__, )
        )

    async def validate(self, attr):
        return attr

    async def run_validation(self, data):
        value = await super(BaseSerializer, self).run_validation(data)
        value = self.validate(value)
        if inspect.isawaitable(value):
            value = await value
        return value

    async def update(self, instance, validated_data):
        raise NotImplementedError('`update()` 必须实现.')

    async def create(self, validated_data):
        raise NotImplementedError('`create()` 必须实现.')

    async def save(self, **kwargs):
        assert hasattr(self, '_errors'), (
            '您必须先调用`.is_valid()`，然后再调用`.save()`。'
        )

        assert not self.errors, (
            '您不能在未通过效验的序列化器上调用`.save()`。'
        )

        # Guard against incorrect use of `serializer.save(commit=False)`
        assert 'commit' not in kwargs, (
            "`commit`不是`save()`方法的有效关键字参数。"
            "如果需要在提交数据库之前访问数据，则"
            "而是检查`serializer.validated_data`."
            "如果您还可以将其他关键字参数传递给`save()`，"
            "需要在保存的模型实例上设置额外的属性。"
            "例如：'serializer.save(owner = request.user)'。"
        )

        assert not hasattr(self, '_data'), (
            '访问`serializer.data`后，您不能调用`.save()`。'
            '如果需要在提交数据库之前访问数据，请访问'
            '`serializer.validated_data`'
        )

        validated_data = dict(list(self.validated_data.items()) + list(kwargs.items()))
        if self.instance is not None:
            validated_data, self.instance = await self.before_update(validated_data, self.instance)
            self.instance = await self.update(self.instance, validated_data)
            assert self.instance is not None, (
                '`update()` 没有返回对象实例。'
            )
        else:
            validated_data, self.instance = await self.before_create(validated_data, self.instance)
            self.instance = await self.create(validated_data)
            assert self.instance is not None, (
                '`create()` 没有返回对象实例。'
            )

        return self.instance

    async def before_update(self, validated_data, instance):
        return validated_data, instance

    async def before_create(self, validated_data, instance):
        return validated_data, instance

    async def is_valid(self, raise_exception=False):
        assert hasattr(self, 'initial_data'), (
            '无法调用`.is_valid()`，因为'
            '类实例化时没有传入`data =`关键字参数'
        )

        if not hasattr(self, '_validated_data'):
            try:
                self._validated_data = await self.run_validation(self.initial_data)
            except ValidationError as exc:
                self._validated_data = {}
                if hasattr(exc, 'error_dict'):
                    self._errors = exc.error_dict
                else:
                    self._errors = {self.field_name, exc.error_list}
            else:
                self._errors = {}

        if self._errors and raise_exception:
            raise ValidationError(self.errors)
        return not bool(self._errors)

    @property
    async def data(self):
        """对外呈现的数据"""
        assert not self.instance is None, '调用 .data 必须先传入 instance= '
        if not hasattr(self, '_data'):
            self._data = await self.internal_to_external(self.instance)
        return self._data

    @property
    def errors(self):
        """对外呈现验证过程中出现的错误"""
        if not hasattr(self, '_errors'):
            msg = '你必须先执行 .is_valid() 方法才能调用 .errors'
            raise AssertionError(msg)
        return self._errors

    @property
    def validated_data(self):
        """对内使用的验证后的数据"""
        if not hasattr(self, '_validated_data'):
            msg = '你必须先执行 .is_valid() 方法才能调用 .validated_data'
            raise AssertionError(msg)
        return self._validated_data


class SerializerMetaclass(type):

    @classmethod
    def _get_declared_fields(cls, bases, attrs):
        fields = [(field_name, attrs.pop(field_name))
                  for field_name, obj in list(attrs.items())
                  if isinstance(obj, Field)]
        fields.sort(key=lambda x: x[1]._sort_counter)

        known = set(attrs)

        def visit(name):
            known.add(name)
            return name

        base_fields = []
        for base in bases:
            if hasattr(base, '_declared_fields'):
                for name, field in base._declared_fields.items():
                    if name not in known:
                        base_fields.append((visit(name), field))

        return OrderedDict(base_fields + fields)

    def __new__(cls, name, bases, attrs):
        attrs['_declared_fields'] = cls._get_declared_fields(bases, attrs)
        return super().__new__(cls, name, bases, attrs)


class Serializer(BaseSerializer, metaclass=SerializerMetaclass):
    default_error_messages = {
        'invalid': '无效数据。应该是字典，但是得到了{datatype}'
    }

    @property
    def fields(self):
        """
        单个格式为 {field_name: field_instance}.
        fields 是动态加载的 避免在导入时出现意想不到的错误
        """
        # like drf
        fields = BindingDict(self)
        for key, value in self.get_fields().items():
            fields[key] = value
        return fields

    @property
    def _writable_fields(self):
        for field in self.fields.values():
            if not field.read_only:
                yield field

    @property
    def _readable_fields(self):
        for field in self.fields.values():
            if not field.write_only:
                yield field

    def get_fields(self) -> dict:
        """
        得到所有fields
        :return:
        """
        return copy.deepcopy(self._declared_fields)

    async def internal_to_external(self, data: Any) -> Any:
        """
        内转外
        :param data:
        :return:

        """
        res = OrderedDict()
        fields = self._readable_fields
        for field in fields:
            try:
                value = await field.get_internal_value(data)
            except SkipField:
                continue
            if value is None:
                res[field.field_name] = None
            else:
                res[field.field_name] = await field.internal_to_external(value)
        return res

    #  反序列化

    async def external_to_internal(self, data: Any) -> Any:
        """
        外转内
        :param data:
        :return:
        """
        res = OrderedDict()
        errors = OrderedDict()
        fields = self._writable_fields
        for field in fields:
            validate_method = getattr(self, 'validate_' + field.field_name, None)
            try:
                primitive_value = field.get_external_value(data)
                validated_value = await field.run_validation(primitive_value)
                if validate_method is not None:
                    validated_value = validate_method(validated_value)
            except ValidationError as exc:
                errors[field.field_name] = exc.error_dict if hasattr(exc, 'error_dict') else exc.error_list
            except SkipField:
                pass
            else:
                set_value(res, field.source_attrs, validated_value)
        if errors:
            raise ValidationError(errors)
        return res

    async def validate(self, attrs):
        return attrs

    def __iter__(self):
        for field in self.fields.values():
            yield self[field.field_name]

    def __getitem__(self, key):
        field = self.fields[key]
        value = self.data.get(key)
        error = self.errors.get(key) if hasattr(self, '_errors') else None
        return {
            'field': field,
            'value': value,
            'error': error,
        }


class ListSerializer(BaseSerializer):
    child = None
    many = True

    default_error_messages = {
        'not_a_list': '预期项目列表，但类型为“ {input_type}”。',
        'empty': '此列表不能为空。'
    }

    def __init__(self, *args, **kwargs):
        self.child = kwargs.pop('child', copy.deepcopy(self.child))
        self.allow_empty = kwargs.pop('allow_empty', True)
        assert self.child is not None, '`child` 是必填参数。'
        assert not inspect.isclass(self.child), '`child` 尚未实例化。'
        super().__init__(*args, **kwargs)
        self.child.bind(field_name='', parent=self)

    async def get_internal_value(self, instance: Any) -> Any:

        for attr in self.source_attrs:
            if isinstance(instance, Mapping):
                instance = instance.get(attr, [])
                if not isinstance(instance, list):
                    instance = [instance]
            else:
                instance = await getattr(instance, attr)
        #         instance = getattr(instance, attr)
        # if inspect.iscoroutine(instance):
        #     instance = await instance
        return instance

    async def internal_to_external(self, data: Any) -> Any:
        """
        内转外
        :param data:
        :return:
        """
        iterable = await data.all() if inspect.isawaitable(data) else data

        return [
            await self.child.internal_to_external(item) for item in iterable
        ]

    async def external_to_internal(self, data: Any) -> Any:
        """
        外转内
        :param data:
        :return:
        """
        if not isinstance(data, list):
            raise self.raise_error('not_a_list', input_type=type(data).__name__)

        if not self.allow_empty and len(data) == 0:
            raise self.raise_error('empty')

        ret = []
        errors = []

        for item in data:
            try:
                value = await self.child.run_validation(item)
            except ValidationError as exc:
                errors.append(exc)
            else:
                ret.append(value)
                errors.append({})
        if any(errors):
            raise ValidationError(errors)
        return ret

    async def run_validation(self, data=empty):
        """
        我们覆盖默认的`run_validation`，因为验证由验证者执行，
        而.validate（）方法应使用“non_fields_error”键被强制为错误字典。
        """
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data
        value = await self.external_to_internal(data)
        self.run_validators(value)
        value = await self.validate(value)
        return value

    async def update(self, instance, validated_data):
        raise NotImplementedError(
            '当 many=True 时，有些序列化器不支持更新操作，'
            '所以在必须使用时请继承ListSerializer并覆盖'
            '`.update()`，不提供默认实现方式')

    async def create(self, validated_data):
        return [
            await self.child.create(attrs) for attrs in validated_data
        ]

    async def save(self, **kwargs):
        """
        保存实例
        """
        # 防止错误使用`serializer.save(commit = False)`
        assert 'commit' not in kwargs, (
            "`commit`不是`save()`方法的有效关键字参数。"
            "如果需要在提交数据库之前访问数据，则"
            "而是检查`serializer.validated_data`."
            "如果您还可以将其他关键字参数传递给`save()`，"
            "需要在保存的模型实例上设置额外的属性。"
            "例如：'serializer.save(owner = request.user)'。"
        )

        validated_data = [
            dict(list(attrs.items()) + list(kwargs.items()))
            for attrs in self.validated_data
        ]

        if self.instance is not None:
            self.instance = await self.update(self.instance, validated_data)
            assert self.instance is not None, (
                '`update()` 没有返回对象实例。'
            )
        else:
            self.instance = await self.create(validated_data)
            assert self.instance is not None, (
                '`create()` 没有返回对象实例。'
            )

        return self.instance


class ModelSerializer(Serializer):
    """
        class Meta:
            model = None
            depth = 0
            extra_kwargs = {}
            fields = ()  # 冲突，不能与exclude共存
            exclude = ()  # 冲突，不能与fields共存
            read_only_fields = ()  # 字段与write_only_fields冲突
            write_only_fields = ()  # 字段与read_only_fields冲突

    """

    @property
    def fields(self):
        """
        单个格式为 {field_name: field_instance}.
        fields 是动态加载的 避免在导入时出现意想不到的错误
        """
        assert hasattr(self, 'Meta'), (
            '{serializer_class} 类没有 "Meta" 属性'.format(
                serializer_class=self.__class__.__name__
            )
        )
        assert hasattr(self.Meta, 'model'), (
            '{serializer_class} 类没有 "Meta.model" 属性'.format(
                serializer_class=self.__class__.__name__
            )
        )
        if self.Meta.model._meta.abstract:
            raise ValueError('不能将ModelSerializer与抽象模型一起使用。')

        declared_fields = copy.deepcopy(self._declared_fields)
        model = getattr(self.Meta, 'model')
        depth = getattr(self.Meta, 'depth', 10)
        if depth is not None:
            assert depth >= 0, "'depth' may not be negative."
            assert depth <= 10, "'depth' may not be greater than 10."

        converter = ModelConverter(ModelSerializer)
        model_original_fields = model._meta.fields_map
        model_clean_fields = self._clean_model_field(model_original_fields)
        effective_field = self.get_effective_field(model_clean_fields)
        serializer_fields = BindingDict(self)

        for field_name, field_class in effective_field.items():
            current_field_class = converter.convert(self, field_class, **self.get_field_kws_by_meta(field_name))
            serializer_fields[field_name] = current_field_class
        serializer_fields.update(declared_fields)
        return serializer_fields

    def get_effective_field(self, model_fields) -> dict:
        """
        得到有效的字段
        :param model_fields: 模型字段
        :return:
        """
        meta_fields = getattr(self.Meta, 'fields', None)
        meta_exclude = getattr(self.Meta, 'exclude', None)
        if meta_exclude and meta_fields:
            raise ValueError('class ”{}“ ’Meta.fields‘ 和 ’Meta.exclude‘ 不可以共存 '.format(self.__class__.__name__))

        if meta_exclude is not None:
            return {k: v for k, v in model_fields.items() if k not in meta_exclude}
        elif meta_exclude is None and (meta_fields is None or meta_fields is ALL_FIELDS):
            return model_fields
        else:
            return {k: v for k, v in model_fields.items() if k in meta_fields}

    def get_field_kws_by_meta(self, field_name):
        """
        判断当前字段是否在Meat内设置了只读或只写
        :param field_name:
        :return:
        """
        read_only_fields = getattr(self.Meta, 'read_only_fields', [])
        write_only_fields = getattr(self.Meta, 'write_only_fields', [])
        if field_name in read_only_fields and field_name in write_only_fields:
            raise ValueError('字段 {} 不可用同时存在于类 ”{}“ 的 ’Meta.read_only_fields‘ '
                             '和 ’Meta.write_only_fields‘ 属性中'.format(field_name, self.__class__.__name__))
        if field_name in read_only_fields:
            return {'read_only': True, 'write_only': False}
        elif field_name in write_only_fields:
            return {'read_only': False, 'write_only': True}
        else:
            return {'read_only': False, 'write_only': False}

    def _clean_model_field(self, model_original_fields):
        """
        清除不需要的字段如 fk_id
        :param model_original_fields:
        :return:
        """
        clean_field_names = []
        field_dict = {}
        fields_map = copy.deepcopy(model_original_fields)
        basis_fields = self._get_model_basis_fields(fields_map)

        for field_name, field_class in basis_fields.items():
            if isinstance(field_class, (tortoise_fields.relational.ForeignKeyFieldInstance, tortoise_fields.relational.OneToOneFieldInstance)):
                clean_field_names.append(field_class.source_field)
            field_dict[field_name] = field_class

        for clean_field_name in clean_field_names:
            if clean_field_name in field_dict:
                field_dict.pop(clean_field_name)
        return field_dict

    async def create(self, validated_data):
        """
        根据验证后的数据进行创建，
        """
        # raise_errors_on_nested_writes('create', self, validated_data)

        ModelClass = self.Meta.model
        ModelClassMeta = ModelClass._meta
        many_to_many = {}

        for m2m_field in ModelClassMeta.m2m_fields:
            if m2m_field in self.fields:
                if m2m_field in validated_data:
                    many_to_many[m2m_field] = validated_data.pop(m2m_field)
        try:
            instance = await ModelClass.create(**validated_data)
        except TypeError as exc:

            raise exc
        if many_to_many:
            for field_name, values in many_to_many.items():
                field = getattr(instance, field_name)
                for value in values:
                    await field.add(value)
        return instance

    async def update(self, instance, validated_data):
        """更新"""
        ModelClass = self.Meta.model
        ModelClassMeta = ModelClass._meta

        m2m_fields = []
        for attr, value in validated_data.items():
            if attr in ModelClassMeta.m2m_fields:
                m2m_fields.append((attr, value))
            else:
                setattr(instance, attr, value)
        await instance.save()
        for attr, values in m2m_fields:
            field = getattr(instance, attr)
            for value in values:
                value, _ = await field.remote_model.get_or_create(**value)
                await field.add(value)
        return instance

    def check_relationship(self):
        """检查关系字段 目前不能为关系字段提供自动转换功能"""

    #
    def _get_model_basis_fields(self, model_fields):
        """
        得到基础字段，非关系字段
        :param model:
        :return:
        """
        return {field_name: field_class for field_name, field_class in model_fields.items() if not isinstance(field_class, tortoise_fields.relational.RelationalField)}
    #
    # def _get_model_M2M_fields(self, model_fields):
    #     """
    #     得到多对多字段
    #     :param model:
    #     :return:
    #     """
    #     return {field_name: field_class for field_name, field_class in model_fields.items() if isinstance(field_class, tortoise_fields.relational.ManyToManyFieldInstance)}
    #
    # def _get_model_O2O_fields(self, model_fields):
    #     """得到一对一字段"""
    #     return {field_name: field_class for field_name, field_class in model_fields.items() if
    #             isinstance(field_class, (tortoise_fields.relational.BackwardOneToOneRelation, tortoise_fields.relational.OneToOneFieldInstance))}
    #
    # def _get_model_M2O_fields(self, model_fields):
    #     """
    #     得到多对一字段
    #     :param model:
    #     :return:
    #     """
    #     return {field_name: field_class for field_name, field_class in model_fields.items() if
    #             isinstance(field_class, tortoise_fields.relational.ForeignKeyFieldInstance)}
    #
    # def _get_model_O2M_fields(self, model_fields):
    #     """
    #     得到一对多字段
    #     :param model:
    #     :return:
    #     """
    #     return {field_name: field_class for field_name, field_class in model_fields.items() if
    #             isinstance(field_class, tortoise_fields.relational.BackwardFKRelation)}



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/3/11 17:37
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    paginations.py
    分页器
@ChangeHistory:
    datetime action why
    example:
    2021/3/11 17:37 change 'Fix bug'
        
"""
from sanic.request import Request
from tortoise import Model

from sanic_rest_framework.exceptions import APIException
from sanic_rest_framework.status import HttpStatus
from sanic_rest_framework.utils import replace_query_param


class BasePagination:
    """抽象基类"""

    async def paginate_queryset(self, queryset, request, view):
        raise NotImplementedError(
            '必须在 `%s` 中实现异步的 `.paginate_queryset()` 方法' % self.__class__.__name__)

    def response(self, request, data):
        raise NotImplementedError(
            '必须在 `%s` 中实现 `.response()` 方法' % self.__class__.__name__)


class GeneralPagination(BasePagination):
    """通用分页器"""
    page_size = 60
    page_query_param = 'page'
    page_size_query_param = 'page_size'
    max_page_size = 10000

    @property
    def count(self):
        assert hasattr(self, '_count'), '必须先执行 `.paginate_queryset()` 函数才能使用.count'
        return self._count

    def get_next_link(self, request: Request):
        assert hasattr(self, '_count'), '必须先执行 `.paginate_queryset()` 函数才能使用.get_next_link()'
        if self.page * self.page_size + self.page_size >= self.count:
            return None
        uri = request.server_path
        query_string = '?' + request.query_string
        query_string = replace_query_param(query_string, self.page_query_param, self.page + 1)
        query_string = replace_query_param(query_string, self.page_size_query_param, self.page_size)
        return uri + query_string

    def get_previous_link(self, request: Request):
        assert hasattr(self, '_count'), '必须先执行 `.paginate_queryset()` 函数才能使用.get_previous_link()'
        if self.page * self.page_size <= 0:
            return None
        uri = request.server_path
        query_string = '?' + request.query_string
        query_string = replace_query_param(query_string, self.page_query_param, self.page - 1)
        query_string = replace_query_param(query_string, self.page_size_query_param, self.page_size)
        return uri + query_string

    async def paginate_queryset(self, queryset, request, view):
        self.page = self.get_query_page(request)
        self.page_size = self.get_query_page_size(request)
        if not isinstance(queryset, Model):
            queryset = queryset.filter()
        self._count = await queryset.count()
        return queryset.limit(self.page_size).offset(self.page * self.page_size)

    def get_query_page(self, request):
        try:
            page = int(request.args.get(self.page_query_param, 0))
        except ValueError as exc:
            raise APIException('发生错误的分页数据', http_status=HttpStatus.HTTP_400_BAD_REQUEST)
        return page

    def get_query_page_size(self, request):
        try:
            page = int(request.args.get(self.page_size_query_param, self.page_size))
            if page > self.max_page_size:
                raise APIException('分页内容大小超出最大限制', http_status=HttpStatus.HTTP_400_BAD_REQUEST)
        except ValueError as exc:
            raise APIException('发生错误的分页数据', http_status=HttpStatus.HTTP_400_BAD_REQUEST)
        return page

    def response(self, request, data):
        return {
            'count': self.count,
            'next': self.get_next_link(request),
            'previous': self.get_previous_link(request),
            'results': data,
        }



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/3/8 11:59
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    test_model_serializers.py
    文件说明
@ChangeHistory:
    datetime action why
    example:
    2021/3/8 11:59 change 'Fix bug'
        
"""
import asyncio
import uuid
from datetime import datetime

from tortoise.contrib.test import initializer, TestCase, finalizer

from sanic_rest_framework.converter import ModelConverter
from sanic_rest_framework.test.models import TestModel, TestOneTOManyModel, TestOneToOneModel, CharEnum, Enum1
from sanic_rest_framework.serializers import ModelSerializer


class TestModelSerializer(ModelSerializer):
    class Meta:
        model = TestModel
        # fields = ('char_field',)
        exclude = ('id',)
        read_only_fields = ('char_field',)
        write_only_fields = ('float_field',)


initializer(['sanic_rest_framework.test.models', ],
            # db_url="sqlite://./db.sqlite",
            loop=asyncio.get_event_loop())


class TestOrdinarySerializer(TestCase):
    def setUp(self) -> None:
        self.data = {
            'char_field': '老四',
            'float_field': 1.36,
            'date_field': '2016-11-1',
            'int_field': 999,
            'decimal_field': 999.9,
            'datetime_field': '2016-12-1 11:1:1',
            'int_enum_field': 1,
            'char_enum_field': '好',
            'boolean_field': False,
            'small_int_field': 1,
            'big_int_field': 99999,
            'text_field': 'PE',
            'json_field': '{"a":1}',
            'uuid_field': '91a4c540-80b5-11eb-b03f-e0d55e47dfb2',
            'one_to_many': [],
            'one_to_one': {},
            'many_to_many': [],
        }

    async def test_serializer(self):
        # for name, field in UserModel._meta.fields_map.items():
        #     ModelConverter().convert(TestModelSerializer, field, name)

        tm = TestModel()
        tm.char_field = 1
        tm.float_field = 1
        tm.date_field = '2016-11-11'
        tm.int_field = 1
        tm.decimal_field = 1
        tm.datetime_field = datetime(2016, 11, 11, 0, 0, 0)
        tm.int_enum_field = Enum1.OK
        tm.char_enum_field = CharEnum.OK
        tm.boolean_field = True
        tm.small_int_field = 1
        tm.big_int_field = 1
        tm.text_field = '1'
        tm.json_field = {}
        tm.uuid_field = str(uuid.uuid1())
        tm.one_to_one = await TestOneToOneModel(name='1').save()
        await tm.save()
        tms = TestModelSerializer(instance=await TestModel.all(), many=True)
        print(await tms.data)
        tms = TestModelSerializer(data=self.data, partial=True)
        tms.is_valid()
        print(tms.validated_data)
        print(tms.fields)

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()



"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/3/11 15:46
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    request.py
    文件说明
@ChangeHistory:
    datetime action why
    example:
    2021/3/11 15:46 change 'Fix bug'
        
"""
from collections import OrderedDict

from sanic.exceptions import InvalidUsage
from sanic.request import Request as SanicRequest


class SRFRequest(SanicRequest):
    def __init__(self, *args, **kwargs):
        super(SRFRequest, self).__init__(*args, **kwargs)
        self.user = None

    @property
    def data(self):
        try:
            data = self.json
        except InvalidUsage as exc:
            data = self.form
        return data



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/19 15:45
@DependencyLibrary: 无
@MainFunction：无
@FileDoc:
    constant.py
    全局常量
"""
ALL_METHOD = {'GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'HEAD', 'OPTIONS'}
DETAIL_METHOD_GROUP = {
    'dynamic_method': ['GET', 'PUT', 'DELETE', 'PATCH'],
    'static_method': ['POST', 'OPTION']
}
LIST_METHOD_GROUP = {
    'dynamic_method': ['PUT', 'DELETE', 'PATCH'],
    'static_method': ['GET', 'POST', 'OPTION']
}



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/19 16:08
@DependencyLibrary: 无
@MainFunction：无
@FileDoc:
    routes.py
    便捷路由文件
"""
from typing import List, Type, Union

from sanic import Sanic, Blueprint

from .constant import ALL_METHOD, DETAIL_METHOD_GROUP, LIST_METHOD_GROUP


# 默认分组


class Route:
    def __init__(self):
        self.routes = []

    def register_route(self, prefix, viewset, name=None):
        """
        注册路由
        :param prefix: url 前缀
        :param viewset: 视图类
        :param name: 供 url_for 使用的名称
        :return:
        """
        if name is None:
            name = prefix

        dynamic_uri = '/{prefix}/<{lookup_field}:string>'
        static_uri = '/{prefix}'
        base_method_group = LIST_METHOD_GROUP
        if hasattr(viewset, 'detail') and viewset.detail:
            base_method_group = DETAIL_METHOD_GROUP

        viewset_methods = self.get_viewset_methods(viewset)
        viewset_dynamic_method = [i for i in viewset_methods if i in base_method_group['dynamic_method']]
        viewset_static_method = [i for i in viewset_methods if i in base_method_group['static_method']]

        if viewset_dynamic_method:
            self.routes.append({
                'handler': viewset.as_view(viewset_dynamic_method),
                'uri': dynamic_uri.format(prefix=prefix, lookup_field=viewset.lookup_field),
                'name': '{name}_detail'.format(name=name)
            })
        if viewset_static_method:
            self.routes.append({
                'handler': viewset.as_view(viewset_static_method),
                'uri': static_uri.format(prefix=prefix),
                'name': '{name}_list'.format(name=name)
            })

    def get_viewset_methods(self, viewset):
        """得到viewSet所有请求方法"""
        methods = []
        for method in ALL_METHOD:
            if hasattr(viewset, method.lower()):
                methods.append(method)
        return methods

    def initialize(self, destination: Union[Sanic, Blueprint]):
        """注册路由"""
        for route in self.routes:
            route['methods'] = ALL_METHOD
            destination.add_route(**route)



from setuptools import find_packages,setup
setup(
    name = 'sanic_rest_framework',
    version = '0.0.1',
    packages = find_packages(),
    description="Sanic rest api framework Similar to DRF ",
    author="WangLaoSi",
    author_email='103745315@qq.com',
    url="https://gitee.com/Wang_LaoSi/sanic_rest_framework",
    download_url='https://gitee.com/Wang_LaoSi/sanic_rest_framework/repository/archive/master.zip',
    install_requires = ['sainc','tortoise-orm']
)


"""
@Author：WangYuXiang
@E-mile：Hill@3io.cc
@CreateTime：2021/2/2 15:58
@DependencyLibrary：无
@MainFunction：无
@FileDoc： 
    test_base_field.py
    字段
@ChangeHistory:
    datetime action why
    example:
    2021/2/2 15:58 change 'Fix bug'
        
"""
import asyncio

from tortoise.contrib import test
from tortoise.contrib.test import initializer, finalizer

from sanic_rest_framework.exceptions import ValidationError
from sanic_rest_framework.fields import Field as TestField, empty, SkipField
from sanic_rest_framework.test.utils import TestDataMixin
from sanic_rest_framework.validators import MaxValueValidator

initializer(['sanic_rest_framework.test.models', ],
            # db_url="sqlite://./db.sqlite",
            loop=asyncio.get_event_loop())


class TestBaseField(TestDataMixin, test.TestCase):
    """测试基类的基本功能"""

    @classmethod
    async def tearDownClass(cls) -> None:
        finalizer()

    def test_bing(self):
        tf1 = TestField()
        tf2 = TestField()
        tf1.bind('test1', tf2)
        self.assertEqual(tf1.field_name, 'test1')
        self.assertEqual(tf1.source, 'test1')
        self.assertEqual(tf1.source_attrs, ['test1'])
        self.assertEqual(tf1.parent, tf2)

    def test_get_external_value(self):
        data1 = {'test': self.str_max_int}
        data2 = {'test': self.max_int}
        data3 = {'test': self.str_chinese}
        data4 = {'test1': self.str_chinese}
        base_tf = TestField()
        tf1 = TestField()
        tf2 = TestField(default=1)
        tf1.bind('test', base_tf)
        tf2.bind('test1', base_tf)

        self.assertEqual(tf1.get_external_value(data1), self.str_max_int)
        self.assertEqual(tf1.get_external_value(data2), self.max_int)
        self.assertEqual(tf1.get_external_value(data3), self.str_chinese)
        self.assertEqual(tf1.get_external_value(data4), empty)
        self.assertEqual(tf1.get_external_value({}), empty)

        self.assertEqual(tf2.get_external_value(data1), 1)
        self.assertEqual(tf2.get_external_value(data2), 1)
        self.assertEqual(tf2.get_external_value(data3), 1)
        self.assertEqual(tf2.get_external_value(data4), self.str_chinese)

    async def test_get_internal_value(self):
        # 未进行 Model 类型测试
        test_data = [
            [self.str_chinese, {'tf': self.str_chinese}],
            [self.str_england, {'tf': self.str_england}],
            [self.max_float, {'tf': self.max_float}],
            [self.str_max_float, {'tf': self.str_max_float}],
            [self.max_int, {'tf': self.max_int}],
            [self.str_max_int, {'tf': self.str_max_int}],
            [self.pi, {'tf': self.pi}],
            [self.str_pi, {'tf': self.str_pi}],
            [self.bool_True, {'tf': self.bool_True}],
            [self.bool_False, {'tf': self.bool_False}],
            [self.str_bool_True, {'tf': self.str_bool_True}],
            [self.str_bool_False, {'tf': self.str_bool_False}],
        ]

        tf = TestField()
        tf.bind('tf', tf)
        for value, data in test_data:
            self.assertEqual(await tf.get_internal_value(data), value)

    def test_run_validators(self):
        tf = TestField(validators=[MaxValueValidator(1000)])
        tf.bind('tf', tf)
        tf.run_validators(10)
        tf.run_validators(1000)
        with self.assertRaises(ValidationError):
            tf.run_validators(1001)

    def test_get_default(self):
        def test():
            return '5'

        tf1 = TestField()
        tf2 = TestField(default=0)
        tf3 = TestField(default=test)
        with self.assertRaises(SkipField):
            tf1.get_default()
        self.assertEqual(tf2.get_default(), 0)
        self.assertEqual(tf3.get_default(), '5')

    def test_validate_empty_values(self):
        base_tf = TestField()

        tf1 = TestField()
        tf1.bind('t1', base_tf)
        self.assertEqual(tf1.validate_empty_values(10), (False, 10))
        with self.assertRaises(SkipField):
            tf1.validate_empty_values(empty)
        with self.assertRaises(ValidationError):
            tf1.validate_empty_values(None)

        tf2 = TestField(default=0)
        tf2.bind('t2', base_tf)
        self.assertEqual(tf2.validate_empty_values(10), (False, 10))
        self.assertEqual(tf2.validate_empty_values(empty), (True, 0))
        with self.assertRaises(ValidationError):
            tf2.validate_empty_values(None)

        tf3 = TestField(default=0, read_only=True)
        tf3.bind('t3', base_tf)
        self.assertEqual(tf3.validate_empty_values(10), (True, 0))
        self.assertEqual(tf3.validate_empty_values(empty), (True, 0))
        self.assertEqual(tf3.validate_empty_values(None), (True, 0))

        tf4 = TestField(default=0, required=True)
        tf4.bind('t4', base_tf)
        self.assertEqual(tf4.validate_empty_values(10), (False, 10))
        with self.assertRaises(ValidationError):
            tf4.validate_empty_values(empty)
        with self.assertRaises(ValidationError):
            tf4.validate_empty_values(None)

        tf5 = TestField(default=0, allow_null=True)
        tf5.bind('t5', base_tf)
        self.assertEqual(tf5.validate_empty_values(10), (False, 10))
        self.assertEqual(tf5.validate_empty_values(empty), (True, 0))
        self.assertEqual(tf5.validate_empty_values(empty), (True, 0))



"""
@Author: WangYuXiang
@E-mile: Hill@3io.cc
@CreateTime: 2021/1/22 22:48
@DependencyLibrary: 无
@MainFunction：无
@FileDoc:
    helpers.py

"""
from collections import OrderedDict, MutableMapping


class BindingDict(MutableMapping):
    """
    这个类似于dict的对象用于在序列化器中存储字段。
    这确保了无论何时将字段添加到我们调用的序列化器中
    field.bind() 使 field_name 和 parent 属性可以正确设置。
    """

    def __init__(self, serializer):
        self.serializer = serializer
        self.fields = OrderedDict()

    def __setitem__(self, key, field):
        self.fields[key] = field
        field.bind(field_name=key, parent=self.serializer)

    def __getitem__(self, key):
        return self.fields[key]

    def __delitem__(self, key):
        del self.fields[key]

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __repr__(self):
        return dict.__repr__(self.fields)
